name = 'ssos'
__version__ = '1.2.3'


GREETING = '''
       *       +      ssos %s
   +        x
        '             https://ssos.readthedocs.io
*   +     _.::'   |
        :_.'    - o - ssos --default : Copy default settings to CWD
|                 |   ssos --help    : Print detailed help
O -     *        '    ssos [-PARAMETER] [-VALUE] path/to/images
|
        _..:'         max.mahlke(at)cab.inta-csic.es
  +   (_.:

        ''' % __version__
